# -*- coding: utf-8 -*-

import math
import random
import cPickle
import hashlib
import shelve

class Vector(object):
    def __init__(self, x = 0., y = 0.):
        self.x = float(x)
        self.y = float(y)

    def __neg__(self):
        return Vector(-self.x, -self.y)

    def __pos__(self):
        return Vector(self.x, self.y)

    def __add__(self, other):
        if isinstance(other, Vector):
            return Vector(self.x + other.x, self.y + other.y)
        elif isinstance(other, (float, int, long)):
            return Vector(self.x + other, self.y + other)
        else:
            raise TypeError('Vector or Number is Required')

    def __sub__(self, other):
        if isinstance(other, Vector):
            return Vector(self.x - other.x, self.y - other.y)
        elif isinstance(other, (float, int, long)):
            return Vector(self.x - other, self.y - other)
        else:
            raise TypeError('Vector or Number is Required')

    def __mul__(self, other):
        if isinstance(other, Vector):
            return Vector(self.x * other.x, self.y * other.y)
        elif isinstance(other, (float, int, long)):
            return Vector(self.x * other, self.y * other)
        else:
            raise TypeError('Vector or Number is Required')

    def __div__(self, other):
        if isinstance(other, Vector):
            return Vector(self.x / other.x, self.y / other.y)
        elif isinstance(other, (float, int, long)):
            return Vector(self.x / other, self.y / other)
        else:
            raise TypeError('Vector or Number is Required')

    def __rmul__(self, other):
        if isinstance(other, Vector):
            return self * other
        elif isinstance(other, (float, int, long)):
            return self * other
        else:
            raise TypeError('Vector or Number is Required')

    def __iadd__(self, other):
        if isinstance(other, Vector):
            self.x += other.x
            self.y += other.y
            return self
        elif isinstance(other, (float, int, long)):
            self.x += other
            self.y += other
            return self
        else:
            raise TypeError('Vector or Number is Required')

    def __isub__(self, other):
        if isinstance(other, Vector):
            self.x -= other.x
            self.y -= other.y
            return self
        elif isinstance(other, (float, int, long)):
            self.x -= other
            self.y -= other
            return self
        else:
            raise TypeError('Vector or Number is Required')

    def __imul__(self, other):
        if isinstance(other, Vector):
            self.x *= other.x
            self.y *= other.y
            return self
        elif isinstance(other, (float, int, long)):
            self.x *= other
            self.y *= other
            return self
        else:
            raise TypeError('Vector or Number is Required')

    def __idiv__(self, other):
        if isinstance(other, Vector):
            self.x /= other.x
            self.y /= other.y
            return self
        elif isinstance(other, (float, int, long)):
            self.x /= other
            self.y /= other
            return self
        else:
            raise TypeError('Vector or Number is Required')

    def __abs__(self):
        return self.length()

    def __str__(self):
        return '[' + str(self.x) + ', ' + str(self.y) + ']'

    def dist(self, other):
        if isinstance(other, Vector):
            return math.sqrt((other.x - self.x) ** 2 + (other.y - self.y) ** 2)
        else:
            raise TypeError('Vector is Required')

    def length(self):
        return math.sqrt(self.x ** 2 + self.y ** 2)

    def normalize(self):
        l = self.length()
        if l > 0.:
            self /= l
        return self

    def dot(self, other):
        if isinstance(other, Vector):
            return self.x * other.x + self.y * other.y
        else:
            raise TypeError('Vector is Required')
        

class Node(object):
    def __init__(self, x, y, color):
        self.r = Vector(x, y)
        self.v = Vector()
        self.c = color

    @property
    def x(self):
        return self.r.x

    @x.setter
    def x(self, x):
        self.r.x = x

    @property
    def y(self):
        return self.r.y

    @y.setter
    def y(self, y):
        self.r.y = y

    @property
    def vx(self):
        return self.v.x

    @vx.setter
    def vx(self, vx):
        self.v.x = vx

    @property
    def vy(self):
        return self.v.y

    @vy.setter
    def vy(self, vy):
        self.v.y = vy

    def dist(self, other):
        if isinstance(other, Node):
            return self.r.dist(other.r)
        elif isinstance(other, Vector):
            return self.r.dist(other)
        else:
            raise TypeError('Node or Vector is Required')

    def delta(self, other):
        if isinstance(other, Node):
            return other.r - self.r
        elif isinstance(other, Vector):
            return other - self.r
        else:
            raise TypeError('Node or Vector is Required')

    def accel(self, a):
        self.v += a

    def update(self):
        self.r += self.v

    def damping(self, damping):
        self.v *= damping

    def restrict(self, region, max_speed):
        speed = self.v.length()
        if speed > max_speed:
            self.v *= max_speed / speed

        if self.x < 0. and self.vx < 0.:
            self.v.x *= -1.
        elif self.x > region[0] and self.vx > 0.:
            self.v.x *= -1.

        if self.y < 0. and self.vy < 0.:
            self.v.y *= -1.
        elif self.y > region[1] and self.vy > 0.:
            self.v.y *= -1.

    def draw_circle(self, ctx, colors):
        ctx.arc(self.x, self.y, 10, 0, 2*math.pi)
        ctx.set_source_rgba(colors[self.c].r, colors[self.c].g, colors[self.c].b, 1.0)
        ctx.fill_preserve()
        ctx.set_source_rgba(1.0, 1.0, 1.0, 1.0)
        ctx.stroke()

class Nodes(object):
    def __init__(self, params):
        self.nodes = []
        self.number = params[0]
        self.region = params[1]
        self.radius = params[2]
        self.force = params[3]
        self.power = params[4]
        self.damping = params[5]
        self.max_speed = params[6]
        self.seed = params[7]

    def append(self, node):
        if isinstance(node, Node):
            self.nodes.append(node)
        else:
            raise TypeError('Node is Required')

    def _attract(self):
        nodes = self.nodes
        radius = self.radius
        force = self.force
        power = self.power
        for i in range(0, len(nodes)):
            for j in range(i+1, len(nodes)):
                delta = nodes[i].delta(nodes[j])
                dist = delta.length()
                if dist > radius or dist < 1.0:
                    continue
                delta.normalize()
                delta *= force * math.pow(dist, power)
                nodes[i].accel(delta)
                nodes[j].accel(-delta)

    def update(self):
        self._attract()

        region = self.region
        max_speed = self.max_speed
        damping = self.damping
        for node in self.nodes:
            node.restrict(region, max_speed)
            node.update()
            node.damping(damping)

    def _draw_line(self, ctx):
        radius = self.radius
        nodes = self.nodes
        
        for i in range(0, len(nodes)):
            for j in range(i+1, len(nodes)):
                dist = nodes[i].dist(nodes[j])

                if dist > radius:
                    continue
                else:
                    ctx.set_source_rgba(1.0, 1.0, 1.0, 1.0 - (dist / radius) ** 2)

                ctx.move_to(nodes[i].x, nodes[i].y)
                ctx.line_to(nodes[j].x, nodes[j].y)
                ctx.stroke()

    def _draw_circle(self, ctx, colors):
        for node in self.nodes:
            node.draw_circle(ctx, colors)

    def draw(self, ctx, colors):
        ctx.set_line_width(2.0)
        self._draw_line(ctx)
        ctx.set_line_width(2.0)
        self._draw_circle(ctx, colors)

#params to hash
def compute_key(params, frame):
    key = cPickle.dumps((params[0], params[1], params[2], params[3], params[4], params[5], params[6], params[7], frame))
    return hashlib.sha1(key).hexdigest()

#get data from db
def get_nodes(d, current_frame, params):
    frame = 0
    nodes = None
    for i in range(current_frame, -1, -1):
        key = compute_key(params, i)
        if d.has_key(key):
            frame = i
            nodes = d[key]
            break

    return frame, nodes

def setup(c, nodes):
    random.seed(int(c.floats[6]))
    for i in range(0, int(c.floats[0])):
        nodes.append(Node(random.random() * c.width, random.random() * c.height, random.randint(0, 3)))

def update(nodes):
    nodes.update()

def draw(ctx, nodes, colors):
    nodes.draw(ctx, colors)

def ompy_params(d):
    d.floatDescs[0].name = 'Number'
    d.floatDescs[1].name = 'Radius'
    d.floatDescs[2].name = 'Force'
    d.floatDescs[3].name = 'Power'
    d.floatDescs[4].name = 'Damping'
    d.floatDescs[5].name = 'Max Speed'
    d.floatDescs[6].name = 'Random Seed'
    d.colorDescs[0].name = 'Color 1'
    d.colorDescs[1].name = 'Color 2'
    d.colorDescs[2].name = 'Color 3'
    d.colorDescs[3].name = 'Color 4'

def ompy_main(c):
    ctx = c.layer_out_context

    d = shelve.open('/ompy/ompy21', protocol = 2)

    params = (0 if c.floats[0] < 0. else int(c.floats[0]),
              (c.width, c.height),
              0. if c.floats[1] < 0. else c.floats[1],
              c.floats[2],
              c.floats[3],
              c.floats[4],
              0. if c.floats[5] < 0. else c.floats[5],
              int(c.floats[6])
             )
    
    frame, nodes = get_nodes(d, c.frame, params)

    if nodes == None:
        nodes = Nodes(params)
        setup(c, nodes)
        key = compute_key(params, 0)
        d[key] = nodes

    for i in range(frame, c.frame):
        update(nodes)
        key = compute_key(params, i + 1)
        d[key] = nodes

    d.close()

    draw(ctx, nodes, c.colors)

