# -*- coding: utf-8 -*-

import csv
import math
import random
import colorsys

class Datum(object):
    def __init__(self, line):
        self.name = line[0]
        self.items = []
        self.left_side = False
        self.angle1 = 0.
        self.angle2 = 0.
        for i in range(1, len(line)):
            self.items.append(float(line[i]))

    def set_leftside(self, left_side):
        self.left_side = left_side

    def _draw_text(self, ctx, center, radius, font_size):
        angle = 0.5 * (self.angle1 + self.angle2)
        e = (math.cos(angle), math.sin(angle))
        radius *= 1.03

        #calculate text width and height
        font_matrix = cairo.Matrix()
        font_matrix.scale(font_size, font_size)
        ctx.set_font_matrix(font_matrix)
        extents = ctx.text_extents(self.name)

        if self.left_side:
            angle += math.pi
            radius += extents[2]

        #rotate text
        font_matrix = cairo.Matrix()
        font_matrix.rotate(angle)
        font_matrix.translate(0, 0.5 * extents[3])
        font_matrix.scale(font_size, font_size)
        ctx.set_font_matrix(font_matrix)
        
        #draw text
        ctx.set_source_rgb(1.0, 1.0, 1.0)
        x = radius * e[0] + center.x
        y = radius * e[1] + center.y
        ctx.move_to(x, y)
        ctx.show_text(self.name)


    def _draw_arc(self, item, ctx, center, radius):
        red, green, blue = colorsys.hls_to_rgb(random.random(), 0.8, 0.8)
        ctx.set_source_rgb(red, green, blue)
        ctx.move_to(center.x, center.y)
        ctx.arc(center.x, center.y, radius, self.angle1, self.angle2)
        ctx.close_path()
        ctx.fill()

    def draw(self, item, accum, sum, ctx, center, radius, phase, max_font_size, time):
        if time == 0.: return 0.
        radius *= time
        self.angle1 = 2. * math.pi * accum / sum - 0.5 * math.pi + phase
        self.angle2 = 2. * math.pi * (accum + self.items[item] * time) / sum - 0.5 * math.pi + phase

        self._draw_arc(item, ctx, center, radius)

        font_size = math.sqrt(self.items[item] / sum) * max_font_size

        #minimum font size
        if font_size < 8.0: font_size = 8.0
        font_size *= time

        self._draw_text(ctx, center, radius, font_size)

        return self.items[item] * time

class Data(object):
    def __init__(self, reader, item, completion, delay):
        self.names = []
        self.data = []

        #get item names
        line = reader.next()
        for i in range(1, len(line)):
            self.names.append(line[i])
        
        for line in reader:
            self.data.append(Datum(line))

        self.item = item % len(self.names)
        datum_num = len(self.data)

        self.completion = completion * 0.01
        self.delay = delay * 0.01
        self.span = 1.0 / (1.0 + (float(datum_num) - 1.0) * self.delay)
        self.interval = self.delay * self.span

        self._sort()
        self._calc_sum()

    def _sort(self):
        self.data.sort(key = lambda x: x.items[self.item], reverse = True)

    def _calc_sum(self):
        item = self.item
        self.sum = 0.
        for datum in self.data:
            self.sum += datum.items[item]

        accum = 0.
        for datum in self.data:
            left_side = (accum + 0.5 * datum.items[item]) / self.sum > 0.5
            datum.set_leftside(left_side)
            accum += datum.items[item]

    def _get_time(self, order):
        time = (self.completion - self.interval * float(order)) / self.span
        return clamp(time, 0., 1.0)

    def _draw_title(self, ctx, title_font_size, title_position):
        ctx.set_source_rgb(1.0, 1.0, 1.0)
        ctx.set_font_matrix(cairo.Matrix())
        ctx.set_font_size(title_font_size)
        ctx.move_to(title_position.x, title_position.y)
        ctx.show_text(self.names[self.item])

    def draw(self, ctx, center, radius, phase, max_font_size, title_font_size, title_position):
        item = self.item
        sum = self.sum
        accum = 0.
        i = 0
        for datum in self.data:
            time = self._get_time(i)
            accum += datum.draw(item, accum, sum, ctx, center, radius, phase, max_font_size, time)
            i += 1

        self._draw_title(ctx, title_font_size, title_position)

def clamp(val, minimum, maximum):
    return max(minimum, min(val, maximum))


def ompy_params(d):
    d.floatDescs[0].name = 'Item'
    d.floatDescs[1].name = 'Completion'
    d.floatDescs[2].name = 'Delay'
    d.floatDescs[3].name = 'Radius'
    d.floatDescs[4].name = 'Phase'
    d.floatDescs[5].name = 'Max Font Size'
    d.floatDescs[6].name = 'Title Font Size'
    d.floatDescs[7].name = 'Color Seed'
    d.pointDescs[0].name = 'Center'
    d.pointDescs[1].name = 'Title Position'

def ompy_main(c):
    ctx = c.layer_out_context
    ctx.set_font_face(cairo.ToyFontFace('Osaka'))

    reader = csv.reader(open('/ompy/ompy24.csv', 'r'))

    item = int(c.floats[0])
    completion = clamp(c.floats[1], 0., 100.)
    delay = clamp(c.floats[2], 0., 100.)

    data = Data(reader, item, completion, delay)

    random.seed(int(c.floats[7]))

    center = c.points[0]
    radius = c.floats[3]
    phase = math.radians(c.floats[4])
    max_font_size = c.floats[5]
    title_font_size = c.floats[6]
    title_position = c.points[1]
    data.draw(ctx, center, radius, phase, max_font_size, title_font_size, title_position)
